require "prey"
require "foraged_cereal"
require "farming"
require "utils"

##
# This class contains the behaviours and attributes for a band of humans.
# Bands follow simple steps of seeking out a new patch, consuming resources and procreating.

class Band

  attr_reader :size, :num_born, :num_died, :num_bands_split

  # In an effort to optimize calculating NARs too many times, we cache previously
  # computed values. This works because the hashcode function in FoodResources takes
  # the resource's size into account, so it invalidates any stale cache entries.
  @@nar_cache = {}

  ##
  # Create a new band with a scape and patch that it belongs and a configuration.
  #
  def initialize(scape, patch, conf)
    @scape = scape
    @current_patch = patch
    @size = size

    @conf = conf # For passing on to children
    # Things we are interested from the passed in conf
    stealables = [:size, :visibility, :search_speed, :search_radius, :search_cost, :handle_cost, :energy_requirement, :birth_chance, :resource_preference]
    stealables.each{|s|self.instance_variable_set("@#{s.to_s}".to_sym, conf[s.to_s])}
    @resource_preference = hash_intern(@resource_preference)
    # This is for logging
    @num_born = 0
    @num_died = 0
    @num_bands_split = 0

    # A precomputed value to make nar computations faster
    @patch_search = (@current_patch.area / (@search_speed * @search_radius * 2))

    if !patch.join
      warn "Warning: Could not join the patch given in the parameter"
      shuffle_about_this_cold_hard_earth
    end
  end

  ##
  # Performs one iteration/epoch, completing each of the steps in the basic behaviour
  # of a band. This is the entry point to getting a band to do something.
  def do_time_step
    shuffle_about_this_cold_hard_earth
    consume
    procreate
  end

  private

  ##
  # Consume as many calories as are needed for one epoch (year). If not enough food
  # is available the population of the band will be decreased.
  def consume
    required_calories = @size * @energy_requirement * 365
    gathered_calories = 0

    # 14-hour work-days, every day of the year, for everybody.
    time_left = 365 * 14

    # Select the resources we are interested in consuming
    resources = select_resources(@current_patch.resources)
    epsilon = 0.00000001 # for floating point comparisons
    while (required_calories - gathered_calories > epsilon and time_left > 0 and
        not @current_patch.depleted?)
      # Get NAR for each resource as well as total NAR
      resource_nars = resources.map do |resource|
        nar(resource)
      end

      # Consume different resources proportional to their NARs
      total_nar = resource_nars.reduce(:+)
      resource_shares = resource_nars.map do |nar|
        (required_calories - gathered_calories) * (nar / total_nar)
      end

      break if total_nar == 0 # Paranoia

      # Search through available resources
      resources.zip(resource_nars, resource_shares) do |resource, nar, target_harvest|
        if nar > 0
          # Either gather as much as needed or as much as the whole group
          # is able to do
          target_harvest = [target_harvest, time_left * nar * @size].min
          actual_harvest = resource.harvest(target_harvest)
          gathered_calories += actual_harvest
          time_left -= actual_harvest / (nar * @size)
        end
      end
    end

    # Contract the population if not enough food was found
    @num_died = 0
    calorie_shortfall = required_calories - gathered_calories
    if calorie_shortfall > epsilon
      @num_died = trim(calorie_shortfall)
    end
  end

  ##
  # Spawn more members, split group into two (or more) if necessary
  # Returns a list of alive bands (0 - infinity)
  def procreate
    @num_born = 0
    @size.times do
      @num_born += 1 if chance(@birth_chance)
    end
    @size += @num_born

    if @size <= 0
      # we died
      @current_patch.leave
      return []
    end
    return [self] if @size < 40 # we're under the split threshold

    # split off bands until we're under the split threshold
    @num_bands_split = 0
    new_bands = [self]
    while @size >= 40
      new_bands << Band.new(@scape, @current_patch, @conf.merge({"size"=>20}))
      @size -= 20
      @num_bands_split += 1
    end
    new_bands
  end

  ##
  # Trim the population of the band that are unfed due to a calorie shortfall
  # Returns the number of band members that were "trimmed"
  def trim(calorie_shortfall)
    stragglers = (calorie_shortfall / (@energy_requirement * 365.0)).ceil
    @size -= stragglers
    stragglers
  end

  ##
  # Searches for, and move to, a new patch
  def shuffle_about_this_cold_hard_earth
    # Need to decide on best patch to move to. Start by getting all visible patches
    visible_patches = @scape.neighbours(@current_patch, @visibility)
    # Now compare patches by "niceness", finding the nicest joinable patch
    memo = [@current_patch, patch_niceness(@current_patch)]
    best_patch, _ = visible_patches.reduce(memo) do |m, patch|
      if patch.joinable?
        niceness = patch_niceness(patch)
        if niceness > m[1]
          m = [patch, niceness]
        end
      end
      m
    end

    # Out with the old, in with the new
    @current_patch.leave
    best_patch.join
    @current_patch = best_patch

    # Recompute the patch search time now that the patch has changed
    @patch_search = (@current_patch.area / (@search_speed * @search_radius * 2))
  end

  ##
  # Calculates nar for the given resource for the current patch
  def nar(resource)
    nar = @@nar_cache[resource] # Check cached nars first
    if nar.nil? # Cache miss
      if !resource.depleted?
        if resource.search_difficulty == 0
          search_time = 0
        else
          search_time = (resource.search_difficulty * @patch_search) / resource.size
        end
        search_penalty = search_time * @search_cost
        handle_penalty = resource.handle_time * @handle_cost
        nar = ((resource.energy_value - search_penalty - handle_penalty) * 60) /
            (resource.handle_time + resource.tend_time + search_time)
        nar = [nar, 0.0].max
      else
        nar = 0.0
      end
      @@nar_cache[resource] = nar
    end
    nar
  end

  ##
  # Decides which resources to use for consumption
  # Mostly used to rule out farming if it's less profitable than foraging
  def select_resources(resources)
    cereals = []
    farms = []
    # We need to do some special attention?
    resources.each do |r|
      if r.is_a?(Cereal)
        cereals << r
      elsif r.is_a?(Farming)
        farms << r
      end
    end

    if farms.size == 0 || cereals.size == 0
      return resources
    end

    if cereals.size > 1
      puts "Bug: more than one cereal resource; can't determine farming switch threshold;
          ignoring all farming resources"
      return resources.select { |r| !r.is_a?(Farming) }
    end
    threshold = nar(cereals.first)
    return resources.select { |r| !r.is_a?(Farming) || nar(r) > threshold }
  end

  ##
  # Weights the nars of each resource by the attractiveness of that resource and sums
  # the values
  def patch_niceness(patch)
    patch.resources.reduce(0.0) do
      |m, o| m + (nar(o) * (@resource_preference[o.name] || 0))
    end
  end

  public
  ##
  # Provides a string representation of the Band. Used for logging.
  def to_s
    "Band (id:#{object_id}) { Patch id: #{@current_patch.object_id} | Size: #{@size} | \
Born/died: #{@num_born}/#{@num_died} | Split: #{@num_bands_split} }"
  end

end
# vim: et:ts=2:sw=2
