##
# Base class for all food resources contained in patches and consumed by bands. Contains a number of
# useful methods for harvesting and replenshing a resource.
# This class on its own is "abstract" and should not be instantiated.

class FoodResource

  attr_reader :size, :energy_value, :handle_time, :capacity, :search_difficulty,
      :repopulation_chance, :name, :tend_time, :carrying_capacity, :colour, :energy_harvested
  # @energy_value defined in subclasses
  # @viable_size defined in subclasses
  # @search_difficulty defined in subclasses

  ##
  # Creates a new food resource instance with a certain carrying capacity (per/km^2) and an area in
  # which the resource exists (km^2).
  # This class on its own is "abstract" and should only be instantiated in subclasses.
  def initialize(cc, area)
    @capacity = @size = cc * area # Resources start at full capacity
    @carrying_capacity = cc
    # Chance per neighbouring healthy patch to repopulate from depletion (percent)
    @repopulation_chance = 10
    @tend_time = 0
    @name = :FoodResource
    @energy_harvested = 0.0
  end

  ##
  # Attempt to harvest this resource for consumption. Harvesters "ask" for a certain number of
  # resources (energy_wanted) and this method returns the number of calories that could be harvested
  # from the resource. It also updates the quantity of the resource accordingly.
  def harvest(energy_wanted)
    quantity_wanted = energy_wanted / @energy_value
    @energy_harvested = 0
    if quantity_wanted > @size
      @energy_harvested = available_energy
      @size = 0
    else
      @energy_harvested = energy_wanted
      @size -= quantity_wanted
    end
    @energy_harvested
  end

  ##
  # If this resource has been entirely consumed then it is depleted.
  def depleted?
    @size <= 0.00001
  end

  ##
  # Returns the total amount of energy available for consumption from this resource.
  def available_energy
    @size * @energy_value
  end

  ##
  # Regrow the resource if it has been depleted. This is generally called by Scape
  def grow
    if @capacity - @size > 0.000001
      exponent = Math.exp(@growth_rate)
      se = @size * exponent
      @size = (se * @capacity) / (@capacity - @size + se)
    end
  end

  ##
  # Logging output
  def to_s
    "Name: #{@name} | Size: #{@size}/#{@capacity}"
  end

  ##
  # Returns true if this resource is dense enough to be what the researchers deemed "viable". Viable
  # patches have a chance of causing depleted resources in neighbouring patches to be repopulated.
  def viable?
    @size > @viable_size
  end

  ##
  # After depletion, resources can be repopulated; i.e. reset their size to the viable size.
  def repopulate
    @size = @viable_size
  end

  ##
  # Returns the percentage of resource available; this is used for patch visualization.
  def percent_full
    @size * @energy_value
  end

  ##
  # The hash code for resources is overloaded so that it is possible to cache pre-calculated NAR
  # values. This works by including the size of the resource in the hash code calculation: if the
  # size changes the cache for the NAR of the resource is invalidated.
  def hash
    [@size, self.object_id].hash
  end

end
# vim: et:ts=2:sw=2
