require 'gtk2'
require 'gui_scape'
require 'gui_graph'

##
# This class encapsulates the overall GUI. It is the main UI window that is displayed when the
# program is started. It also manages timing and coordination between the visualization and the
# underlying simulation.
class GUI < Gtk::Window

  ##
  # Initializes a GUI instance with a simulation backing it. The window is shown.
  def initialize(simulation)
    super()
    @simulation = simulation
    @time = Time.now

    # Close button quits application
    signal_connect("destroy") do
      Gtk.main_quit
      false
    end

    set_title "From Hunter-gatherers to Agriculture"
    set_border_width 10
    set_default_size(640, 480)

    @notebook = create_notebook # tab interface

    # Timing state
    @running = false
    @speed = 15
    @progress = 0
    @iterations = 100
    @last_sim = Time.now

    reset_gui
  end

  ##
  # Resets all the user interface elements to the state they were when the application was started.
  def reset_gui
    each { |c| remove(c) } # Remove all child widgets
    @gui_scape = GUIScape.new(@simulation.scape)
    ui = create_root_box(create_controls, @gui_scape)
    @population_graph = GUIGraph.new
    @resources_graph = GUIGraph.new

    @notebook.append_page(ui, Gtk::Label.new("Simulation"))
    @notebook.append_page(@population_graph, Gtk::Label.new("Population graph"))
    @notebook.append_page(@resources_graph, Gtk::Label.new("Resources graph"))
    @notebook.set_page(0)

    add(@notebook)
    show_all
  end

  ##
  # Begins the main Gtk loop. This is a blocking call and the Gtk loop will take control of the
  # thread it is called from.
  def gtk_main
    Gtk.main
  end

  private

  ##
  # Create and return the tab interface with tab callback attached
  def create_notebook
    notebook = Gtk::Notebook.new
    notebook.signal_connect("switch_page") { |s, p, n| on_page_switch(s, p, n) }
    notebook
  end

  ##
  # Create and return the main layout for the simulation page
  def create_root_box(controls, scape)
    root_box = Gtk::HBox.new(false, 10)

    # Add buttons on the left and grid on the right
    root_box.pack_start(controls, false, false)
    root_box.pack_start(scape, true, true)
  end

  ##
  # Create and return all the controls in a Gtk::VBox
  def create_controls
    controls_box = Gtk::VBox.new(false, 10)
    controls_box.pack_start(speed_control, false)
    controls_box.pack_start(Gtk::HSeparator.new)
    controls_box.pack_start(iterations_control, false)
    controls_box.pack_start(Gtk::HSeparator.new)
    bar_types.each{|w|controls_box.pack_start(w,false)}
    controls_box.pack_start(Gtk::HSeparator.new)
    progress_bar.each{|w|controls_box.pack_start(w,false)}
    controls_box.pack_start(Gtk::HSeparator.new)
    controls_box.pack_start(reset_button,false)
    controls_box.pack_start(start_stop_button, false)
  end

  ##
  # Create the "Turns per second control"
  def speed_control
    @speed_scale = Gtk::HScale.new(1, 1000, 1)
    @speed_scale.signal_connect("value-changed") { |w| on_speed_changed(w) }
    @speed_scale.value = @speed
    add_label(@speed_scale, "Turns per second:")
  end

  ##
  # Create the control for determining the number of iterations to run
  def iterations_control
    @iterations_spin_button = Gtk::SpinButton.new(1, 10000, 10)
    @iterations_spin_button.signal_connect("value-changed") { |w| on_iterations_changed(w) }
    @iterations_spin_button.set_value(@iterations)
    add_label(@iterations_spin_button, "Simulation iterations:")
  end

  ##
  # Create the controls for changing the bar visualization
  def bar_types
      capacity = Gtk::RadioButton.new("Percentage of capacity")
      energy =Gtk::RadioButton.new(capacity, "Energy capacity")
      size = Gtk::RadioButton.new(capacity, "Size")

      capacity.signal_connect("toggled") do
        @gui_scape.config_update(:bar_type => :capacity,
            :bar_proc => Proc.new { |r| r.size / r.capacity })
      end
      energy.signal_connect("toggled") do
        @gui_scape.config_update(:bar_type => :energy,
            :bar_proc => Proc.new { |r| r.size * r.energy_value })
      end
      size.signal_connect("toggled") do
        @gui_scape.config_update(:bar_type => :size, :bar_proc => Proc.new { |r| r.size })
      end
      capacity.set_active(true)
      [Gtk::Label.new("Graph type:"),capacity, energy, size]
  end

  ##
  # Create the progress bar and the iterations counter
  def progress_bar
    @progress_bar = Gtk::ProgressBar.new
    @progress_bar.set_pulse_step(0.1)
    @progress_bar.set_fraction(@progress / @iterations)
    @iterations_counter = Gtk::Label.new("Iteration #{@progress}")
    [@iterations_counter, @progress_bar]
  end

  ##
  # Create the start/stop/play/pause button
  def start_stop_button
    @start_button = Gtk::Button.new("Start simulation")
    @start_button.signal_connect("clicked") { |w| on_start_stop_clicked(w) }
    @start_button
  end

  ##
  # Create the button to reset the simulation
  def reset_button
    @reset_button = Gtk::Button.new("Reset")
    @reset_button.signal_connect("clicked") { |w| @reset = true}
    @reset_button
  end

  ##
  # Utility method for adding a label above a widget. Widget and label are wrapped in a Gtk::VBox
  def add_label(widget, label_text)
    box = Gtk::VBox.new(false, 5)
    label = Gtk::Label.new(label_text)
    box.pack_start(label, false)
    box.pack_start(widget, false)
  end

  ##
  # Callback for the start/stop button
  def on_start_stop_clicked(widget)
    @running ? stop : start
  end

  ##
  # Starts the simulation and visualization
  def start
    @running = true
    @simulation.start_logging
    @simulation.start_graph_caching
    @gui_update_id = Gtk.timeout_add(1000 / 20) { update_gui }
    @simulation_update_id = Gtk.idle_add{ update_simulation }

    @start_button.label = "Stop simulation"
  end

  ##
  # Stops the simulation and visualization
  def stop
    @running = false
    Gtk.timeout_remove(@gui_update_id)
    Gtk.idle_remove(@simulation_update_id)

    @start_button.label = "Start simulation"
    false
  end

  ##
  # Resets the visualization and simulation
  def reset
    stop
    @simulation.reset
    @progress = 0
    reset_gui
  end

  ##
  # Callback for the speed slider widget
  def on_speed_changed(widget)
      @speed = widget.value
  end

  ##
  # Callback for the "turns per second" spin box widget
  def on_iterations_changed(widget)
    if widget.value < @progress
      @iterations = @progress
      @iterations_spin_button.set_value(@iterations)
      @progress_bar.set_fraction(1)
      return
    end
    @iterations = widget.value
    @progress_bar.set_fraction(@progress/@iterations) if !@progress_bar.nil?
  end

  ##
  # Callback for notebook page changes. Updates graphs if necessary.
  # Page is not useful. Not fully implemented on Ruby/Gtk2.
  def on_page_switch(notebook, page, page_num)
    return if !@simulation.graph_caching_enabled # Nothing to do if graphs off

    case page_num
    when 0 # Simulation
    when 1 # Population graph
      @population_graph.show_graph(@simulation.create_population_graph, @simulation.epoch)
    when 2 # Resources graph
      @resources_graph.show_graph(@simulation.create_resources_graph, @simulation.epoch)
    end
  end

  ##
  # Called from the idle handler in the Gtk thread. Performs the simulation update tick.
  # This method returns false if it should not be called again by the idle handler.
  def update_simulation
    if @reset
      @reset = false
      reset
      return false
    end
    return stop if @progress >= @iterations
    sleep([@last_sim + 1 / @speed - Time.now, 0].max)
    begin
      @simulation.do_time_step
      @last_sim = Time.now
      @progress += 1
    rescue StandardError => e # Gtk "hides" exceptions from us so we must print here
      puts e.message
      puts e.backtrace
      return false
    end
    true
  end

  ##
  # Update the GUI. This is called from the Gtk::timeout mechanism so the GUI thread sleeps between
  # the scheduled calls to this method.
  def update_gui
    begin
       @gui_scape.update
       @iterations_counter.text = "Iteration #{@progress}"
       @progress_bar.set_fraction(@progress / @iterations)
    rescue Error => e
      puts e.message
      puts e.backtrace
      return false
    end
    true
  end

end
# vim: et:ts=2:sw=2
