require 'gtk2'

##
# GUIPatche instances are each linked to specific Patch instances and visualize the current resource
# quantities as well as the number of bands present in the patch.
class GUIPatch < Gtk::DrawingArea

  # Colours for mini bar graph
  MAX_SHADE = 0.8
  MIN_SHADE = 0.1

  # Colours for agent count
  COLOR_AGENT_CIRCLE = [1.0, 1.0, 0.0, 0.6] # Mostly-transparent yellow
  COLOR_AGENT_TEXT = [0.0, 0.0, 0.0, 1.0] # Black

  # Size in pixels that the cell will request to be
  WIDTH = 21
  HEIGHT = 21

  # Accessors
  attr_reader :prey
  attr_reader :cereal
  attr_reader :farm
  attr_reader :bands

  ##
  # Initialize a GUIPatch to track a certain patch. 'Shade' specifies the percent of grey to shade
  # the widget's background.
  def initialize(patch, shade)
    super()
    @patch = patch
    shade_range = MAX_SHADE - MIN_SHADE
    @shade = Array.new(3) { MAX_SHADE - shade * shade_range }

    # The maximum value, for vertical bar scaling
    @energy_scale = patch.resources.map{|r|r.size * r.energy_value}.max
    @size_scale = patch.resources.map{|r|r.capacity}.max

    # bar_proc derives a metric for the given resource to show on the bar graph
    # This is a default one; it will get overriden by the default radio button selection
    @bar_proc = Proc.new { |r| r.size / r.capacity }
    # This is a default scaling factor for the numbers the bar_proc spits out
    # As with the bar_proc, it will be overriden using the config_update method before any actual
    # execution
    @scale = 1

    # Initialize stores for stats
    @percentages = {}
    @bands = 0

    # Request a size and connect the draw call
    set_size_request(WIDTH, HEIGHT)
    signal_connect("expose_event") { draw }

    update
  end

  ##
  # Updates the configuration of the patch. Sets which patch value to use as the scale for the bars.
  def config_update(conf)
    if bar_type = conf[:bar_type]
      case bar_type
      when :capacity
        @scale = 1
      when :energy
        @scale = @energy_scale
      when :size
        @scale = @size_scale
      end
    end
    if bar_proc = conf[:bar_proc]
      @bar_proc = bar_proc
    end
  end

  ##
  # Update the bar graph. Alternatively can set quantities individually through accessors.
  # Prey, cereal and farm are floats between 0 and 1 (percentages)
  #
  def update
    redraw_necessary = false
    @patch.resources.each do |r|
      new_val = @bar_proc.call(r) / @scale
      redraw_necessary = true if !redraw_necessary && @percentages[r] != new_val
      @percentages[r] = new_val
    end
    @bands = @patch.num_bands

    queue_draw if redraw_necessary
  end

  private

  ##
  # Draw the patch using Cairo vector operations.
  def draw
    # Use cairo for drawing vector graphics
    cairo = window.create_cairo_context

    # TODO change to use the max across all patches
    num_resources = @patch.resources.count

    # All bars will be a proportion of the width and some portion of the height of the view
    width = allocation.width / num_resources
    height = allocation.height

    # Draw background shade
    cairo.set_source_rgb(@shade)
    cairo.rectangle(0, 0, allocation.width, allocation.height)
    cairo.fill

    # Draw resource bars
    @patch.resources.each_with_index do |r,i|
      value = @percentages[r]
      next if value.nil?
      x = width * i
      y = height * (1.0 - value)
      cairo.set_source_rgb(r.colour)
      cairo.rectangle(x,y,width,height-y)
      cairo.fill
    end

    # Bands/agents
    if (@bands > 0)
      # First the shadow circle
      x = allocation.width / 2
      y = allocation.height / 2
      cairo.set_source_color(COLOR_AGENT_CIRCLE)
      cairo.circle(x, y, [allocation.width/3,allocation.height/3].min)
      cairo.fill

      # Now the count
      cairo.set_source_color(COLOR_AGENT_TEXT)
      cairo.move_to(x - 2 - Math.log(@bands,10).floor*5,y + 4) # Determined by exact science
      cairo.show_text @bands.to_s
    end
  end

end
# vim: et:ts=2:sw=2
