require 'patch'
require 'band'
require 'scape'
require 'prey'
require 'cereal'
require 'yaml'
require 'kwalify' # External gem
require 'utils'
require 'errors'
require 'time'
require 'bounded_queue'
require 'gruff'

##
# Class that contains the overall simulation. Manages the scape and the bands.

class Simulation

  attr_reader :scape, :epoch, :logging_enabled, :graph_caching_enabled

  GRAPH_CACHE_SIZE = 1000 # Number of datapoints cached for graphs

  ##
  # Initializes a new simulation. The config file is parsed and parameters are passed on to the
  # other objects that make up the simulation.
  def initialize(config_file)
    supposed_config = generate_config(config_file)
    begin
      @config = YAML.load(supposed_config)
    rescue CustomErrors::InvalidConfigError
      warn "Error: invalid config; dumping to invalid.cfg"
      File.open("invalid.cfg") { |f| f.write(supposed_config) }
      raise "Invalid config"
    end
    reset
  end

  ##
  # Resets the simulation to its start state
  def reset
    @scape, @bands = parse_config(@config)
    @epoch = 0
    @logging_enabled = false
    @graph_caching_enabled = false
  end

  ##
  # Performs one epoch (year) of the simulation. Subsequent calls of 'do_time_step' are triggered
  # on the scape and the bands. Logging and graph statistics caching also happen here if they have
  # been started.
  def do_time_step
    log if @logging_enabled
    cache_graph_data if @graph_caching_enabled

    @epoch += 1
    @scape.do_time_step
    new_bands = []
    @bands.each do |b|
      new_bands += b.do_time_step
    end
    @bands = new_bands
  end

  private

  ##
  # Given a filename of a config, produces the actual config string
  # Executes files with a hashbang, and uses the stdout as the config
  # Otherwise, reads in the file
  def generate_config(config_file)
    if File.readlines(config_file).first =~ /#!.*/
      begin
        # If file has a hashbang, execute it, and treat stdout as config
        return `#{File.expand_path(config_file)}`
      rescue Errno::EACCES # File is not executable
      warn "Warning: #{config_file} had a hashbang, but was not executable; treating literally"
      end
    end
    return File.read(config_file) # Default to treating file itself as config
  end

  ##
  # Uses the Kwalify gem to validate the generated file with our schema
  def validate_config(config)
    metavalidator = Kwalify::MetaValidator.instance
    begin
      errors = metavalidator.validate(YAML.load(File.read('schema.yml')))
      if !errors.empty?
        errors.each{|e| puts e.message}
        raise CustomErrors::InvalidSchemaError
      end
    rescue Kwalify::SyntaxError, Psych::SyntaxError => e
      puts e.message
      raise CustomErrors::InvalidSchemaError
    end
    schema = Kwalify::Yaml.load_file('schema.yml')
    validator = Kwalify::Validator.new(schema)
    begin
      errors = validator.validate(config)
    rescue
      raise CustomErrors::InvalidConfigError
    end
    if !errors.empty?
      errors.each{|e|puts e}
      raise CustomErrors::InvalidConfigError
    end
  end

  ##
  # Parses the config object; returns pair of scape, band array
  def parse_config(config)
    validate_config(config)
    bands = []
    patches = config["patches"].map do |row|
      row.map do |patch|
        resources = hash_intern(patch["resources"])
        p = Patch.new(resources, patch)
        patch["bands"].each do |band|
          bands << [p,band] # We will need the scape for initialisation; store the parameters for now
        end
        p
      end
    end
    scape = Scape.new(patches)
    bands.map!{|b| Band.new(scape,b[0],b[1])}
    [scape, bands]
  end

  public

  ##
  # Initializes and enables graph statistics caching. This is necessary if graphs are desired.
  # A folder called "graphs" will be created if one does not already exist. Graphs will be stored
  # there.
  def start_graph_caching
    return if @graph_caching_enabled

    Dir.mkdir("graphs") unless File.exists?("graphs")

    @population_cache = BoundedQueue.new(GRAPH_CACHE_SIZE) # Integers
    @num_deaths_cache = BoundedQueue.new(GRAPH_CACHE_SIZE) # Integers

    @resource_consumption_cache = BoundedQueue.new(GRAPH_CACHE_SIZE) # string : float maps

    @graph_caching_enabled = true
  end

  ##
  # Creates a graph of the current population and saves it with a timestamped filename to the
  # "graphs" folder. The filepath is returned.
  def create_population_graph
    graph = Gruff::Line.new
    graph.title = "Population"
    graph.y_axis_label = "Population"
    graph.x_axis_label = "Epochs"
    graph.labels = create_graph_labels
    graph.data(:Population, @population_cache)
    graph.data(:Deaths, @num_deaths_cache)
    write_graph_file(graph, "population")
  end

  ##
  # Creates a graph of the current resource consumption and saves it with a timestamped filename to
  # the "graphs" folder. The filepath is returned.
  def create_resources_graph
    graph = Gruff::Line.new
    graph.title = "Resource consumption"
    graph.y_axis_label = "Calories consumed"
    graph.x_axis_label = "Epochs"
    graph.labels = create_graph_labels
    data = {}
    @resource_consumption_cache.each do |entry|
      entry.each do |k, v|
        data[k] = data.fetch(k, []) << v
      end
    end
    data.each { |k, v| graph.data(k, v) }
    graph.marker_font_size = 16
    write_graph_file(graph, "resources")
  end

  private

  ##
  # Creates graph x-axis labels
  def create_graph_labels
    labels = {}
    if @epoch < GRAPH_CACHE_SIZE
      start = 0
      steps = @epoch / 10
    else
      start = @epoch - GRAPH_CACHE_SIZE
      steps = GRAPH_CACHE_SIZE / 10
    end
    10.times { |t| labels[t * steps] = (steps * t + start).to_s }
    labels
  end

  ##
  # Write a graph image file to disk with a timestamp
  def write_graph_file(graph, name)
    time = Time.now.strftime("%Y-%m-%dT-%T")
    filepath = "graphs/#{name}-#{time}.png"
    graph.write(filepath)
    filepath
  end

  ##
  # Cache data for the graphs to the bounded queue caches.
  def cache_graph_data
    @population_cache.push(population)
    @num_deaths_cache.push(number_died)
    @resource_consumption_cache.push(@scape.resource_consumption)
  end

  public

  ##
  # Initialize and enable logging for the bands and the scape. A folder, "logs", will be created if
  # one does not already exist. Logfiles are timestamped text files.
  def start_logging
    return if @logging_enabled

    Dir.mkdir("logs") unless File.exists?("logs")
    @bands_logfile = open_logfile("bands") if @bands_logfile.nil?
    @scape_logfile = open_logfile("scape") if @scape_logfile.nil?

    @logging_enabled = true
  end

  private

  ##
  # Save logging data to the log files for bands and scapes.
  def log
    @bands.each { |band| @bands_logfile << band.to_s << "\n" }
    @scape_logfile << @scape.to_s
  end

  ##
  # Create a logfile with a timestamp and the given name.
  def open_logfile(name)
    time = Time.now.strftime("%Y-%m-%dT-%T")
    File.open("logs/#{name}-#{time}.log", "w")
  end

  ##
  # Calculate the current total population.
  def population
    @bands.reduce(0) { |m, band| m + band.size }
  end

  ##
  # Calculate the total number of agents born this iteration.
  def number_born
    @bands.reduce(0) { |m, band| m + band.num_born }
  end

  ##
  # Calculate the total number of agents that died this iteration.
  def number_died
    @bands.reduce(0) { |m, band| m + band.num_died }
  end

end
# vim: et:ts=2:sw=2
